# -*- coding: utf-8 -*-
# pyxdsm is a python library for generating publication quality PDF XDSM
# diagrams.This library is a thin wrapper that uses the TIKZ library and
# LaTeX to build the PDFs. Source: https://github.com/mdolab/pyXDSM
#
# OpenMDAO Open Source License:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this software except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division

from collections import namedtuple
import os

import numpy as np


# color pallette link: http://paletton.com/#uid=72Q1j0kllllkS5tKC9H96KClOKC
base_file_start = r"""\documentclass[border=0pt]{standalone}
% Justin S. Gray 2018
% Based off code by John T. Hwang (2014), who based his code off Alec Jacobson: http://www.alecjacobson.com/weblog/?p=1289

% nc = necessary comment [do not remove]

% Four rules for using these macros:
% 1. Always start with a row
% 2. Rows contain cols and cols contain rows
% 3. Mats should be on at least the 3rd level (row->col->mat, minimum)
% 4. If a row contains the mat, add &; if a col contains the mat, add \\

% ---------------------------------------

\usepackage{tikz}
\usepackage{ifthen}
\usepackage{esdiff}
\usepackage{varwidth}

\definecolor{tableau0}{RGB}{77, 121, 168}
\definecolor{tableau1}{RGB}{242, 142, 43}
\definecolor{tableau2}{RGB}{255, 87, 88}
\definecolor{tableau3}{RGB}{118, 183, 178}
\definecolor{tableau4}{RGB}{89, 161, 78}
\definecolor{tableau5}{RGB}{237, 201, 72}
\definecolor{tableau6}{RGB}{176, 121, 162}
\definecolor{tableau7}{RGB}{255, 157, 167}
\definecolor{tableau8}{RGB}{156, 116, 94}
\definecolor{tableau9}{RGB}{186, 176, 172}


\newcommand{\thk}{0.01in}
\newcommand{\thkln}{0.02in}

% \blockmat{width}{height}{text}{block_options}{other}
\newcommand{\blockmat}[5]{
\begin{tikzpicture}
    \draw[draw=white,fill=white,#4,line width=\thk] (0,0) rectangle( #1-\thk,#2-\thk);
    #5
    \draw (#1/2, #2/2) node {#3};
\end{tikzpicture}
}

% blockempty{width}{height}{text}
\newcommand{\blockempty}[3]{
    \blockmat{#1}{#2}{#3}{draw=white,fill=white}{}%
}

% \blockmat{width}{height}{text}{block_options}{diagonal_width}{diagonal_options}
\newcommand{\blockdiag}[6]{
    \blockmat{#1}{#2}{#3}{#4}
    {
    \draw[#6,line width=\thk] (0,#2-\thk) -- (#5,#2-\thk) -- ( #1-\thk,#5) -- ( #1-\thk,0) -- ( #1-\thk - #5,0) -- (0,#2-\thk -#5) --cycle;
    }%
}

% \blockddots{width}{height}{text}{block_options}{dot_radius}{dot_options}{dot_h}{dot_v}
\newcommand{\blockdots}[8]{
    \blockmat{#1}{#2}{#3}{#4}%
    {%
    \ifthenelse{\equal{#5}{}}
    {\newcommand\dotradius{0.01in}}
    {\newcommand\dotradius{#5}}%
    \filldraw[#6] (#1/2, #2/2) circle (0.5*\dotradius);%
    \filldraw[#6] (#1/2 + #7, #2/2 + #8) circle (0.5*\dotradius);%
    \filldraw[#6] (#1/2 - #7, #2/2 - #8) circle (0.5*\dotradius);%
    }%
}

% \leftbracket{width}{height}{options}
\newcommand{\leftbracket}[3]{
    \begin{tikzpicture}
        \coordinate (iSW) at (\thk+\thkln/2,\thk+\thkln/2);
        \coordinate (iNW) at (\thk+\thkln/2,#2-\thk-\thkln/2);
        \coordinate (iSE) at (#1-\thk-\thkln/2,\thk+\thkln/2);
        \coordinate (iNE) at (#1-\thk-\thkln/2,#2-\thk-\thkln/2);
        \coordinate (oSW) at (\thk/2,\thk/2);
        \coordinate (oNW) at (\thk/2,#2-\thk/2);
        \coordinate (oSE) at (#1-\thk/2,\thk/2);
        \coordinate (oNE) at (#1-\thk/2,#2-\thk/2);
        \draw[#3,line width=\thkln] (iNE) -- (iNW) -- (iSW) -- (iSE);
        \draw[draw=white,line width=\thk] (oNE) -- (oNW) -- (oSW) -- (oSE);
    \end{tikzpicture}%nc
}

% \rightbracket{width}{height}{options}
\newcommand{\rightbracket}[3]{
    \begin{tikzpicture}
        \coordinate (iSW) at (\thk+\thkln/2,\thk+\thkln/2);
        \coordinate (iNW) at (\thk+\thkln/2,#2-\thk-\thkln/2);
        \coordinate (iSE) at (#1-\thk-\thkln/2,\thk+\thkln/2);
        \coordinate (iNE) at (#1-\thk-\thkln/2,#2-\thk-\thkln/2);
        \coordinate (oSW) at (\thk/2,\thk/2);
        \coordinate (oNW) at (\thk/2,#2-\thk/2);
        \coordinate (oSE) at (#1-\thk/2,\thk/2);
        \coordinate (oNE) at (#1-\thk/2,#2-\thk/2);
        \draw[#3,line width=\thkln] (iNW) -- (iNE) -- (iSE) -- (iSW);
        \draw[draw=white,line width=\thk] (oNW) -- (oNE) -- (oSE) -- (oSW);
    \end{tikzpicture}%nc
}

% \upperbracket{width}{height}{options}
\newcommand{\upperbracket}[3]{
    \begin{tikzpicture}
        \coordinate (iSW) at (\thk+\thkln/2,\thk+\thkln/2);
        \coordinate (iNW) at (\thk+\thkln/2,#2-\thk-\thkln/2);
        \coordinate (iSE) at (#1-\thk-\thkln/2,\thk+\thkln/2);
        \coordinate (iNE) at (#1-\thk-\thkln/2,#2-\thk-\thkln/2);
        \coordinate (oSW) at (\thk/2,\thk/2);
        \coordinate (oNW) at (\thk/2,#2-\thk/2);
        \coordinate (oSE) at (#1-\thk/2,\thk/2);
        \coordinate (oNE) at (#1-\thk/2,#2-\thk/2);
        \draw[#3,line width=\thkln] (iSW) -- (iNW) -- (iNE) -- (iSE);
        \draw[draw=white,line width=\thk] (oSW) -- (oNW) -- (oNE) -- (oSE);
    \end{tikzpicture}%nc
}

% \lowerbracket{width}{height}{options}
\newcommand{\lowerbracket}[3]{
    \begin{tikzpicture}
        \coordinate (iSW) at (\thk+\thkln/2,\thk+\thkln/2);
        \coordinate (iNW) at (\thk+\thkln/2,#2-\thk-\thkln/2);
        \coordinate (iSE) at (#1-\thk-\thkln/2,\thk+\thkln/2);
        \coordinate (iNE) at (#1-\thk-\thkln/2,#2-\thk-\thkln/2);
        \coordinate (oSW) at (\thk/2,\thk/2);
        \coordinate (oNW) at (\thk/2,#2-\thk/2);
        \coordinate (oSE) at (#1-\thk/2,\thk/2);
        \coordinate (oNE) at (#1-\thk/2,#2-\thk/2);
        \draw[#3,line width=\thkln] (iNW) -- (iSW) -- (iSE) -- (iNE);
        \draw[draw=white,line width=\thk] (oNW) -- (oSW) -- (oSE) -- (oNE);
    \end{tikzpicture}%nc
}

% a hack so that I don't have to worry about the number of columns or
% spaces between columns in the tabular environment
\newenvironment{blockmatrixtabular}
{%nc
    \renewcommand{\arraystretch}{0}%nc
    \begin{tabular}{
    @{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l
    @{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l
    @{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l@{}l
    @{}
    }%nc
}
{
    \end{tabular}%nc
}

% \blockcol{
% }
\newcommand{\blockcol}[1]{\vtop{\null\hbox{\begin{blockmatrixtabular}#1\end{blockmatrixtabular}}}&}

% \blockrow{
% }
\newcommand{\blockrow}[1]{\begin{blockmatrixtabular}#1\end{blockmatrixtabular}\\}


\begin{document}
\begin{varwidth}{10\textwidth}

\newcommand\mwid{0.5in}
\newcommand\wid{0.15in}
\newcommand\comp{0.3in}
\newcommand\ext{0.5in}
\newcommand\dt{0.03in}
\newcommand\txt{0.8in}

\definecolor{Tgrey}{rgb}{0.9,0.9,0.9}
\definecolor{Tred}{rgb}{1.0,0.722,0.714}
\definecolor{Tgreen}{rgb}{0.639,0.89,0.655}
\definecolor{Tblue}{rgb}{0.667,0.631,0.843}
\definecolor{Tyellow}{rgb}{1,0.941,0.714}

\definecolor{Lred}{rgb}{17.3,0.063,0.059}
\definecolor{Lgreen}{rgb}{0.047,0.133,0.051}
\definecolor{Lblue}{rgb}{0.063,0.051,0.118}
\definecolor{Lyellow}{rgb}{0.173,0.149,0.059}

\definecolor{Dgrey}{rgb}{0.4,0.4,0.4}
\definecolor{Dred}{rgb}{1.0,0.333,0.318}
\definecolor{Dgreen}{rgb}{0.282,0.89,0.322}
\definecolor{Dblue}{rgb}{0.42,0.341,0.843}
\definecolor{Dyellow}{rgb}{1.0,0.863,0.318}

\definecolor{Bred}{rgb}{0.302,0.8,0.0}
\definecolor{Bgreen}{rgb}{0.4,1.0,0.4}
\definecolor{Bblue}{rgb}{0.043,0.012,0.208}
\definecolor{Byellow}{rgb}{0.302,0.243,0.0}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
"""
base_file_end = r"""
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\end{varwidth}
\end{document}"""


Variable = namedtuple('Variable', field_names=['size', 'idx', 'text', 'color'])

CellData = namedtuple('CellData', field_names=['text', 'color', 'highlight'])


def _color(base_color, h_light):
    if h_light == -1:
        color = 'white'
    elif h_light == 0:
        color = 'Tgrey'
    elif h_light == 1:
        color = 'T{}'.format(base_color)
    elif h_light == 2:
        color = 'D{}'.format(base_color)
    elif h_light == 3:
        color = 'B{}'.format(base_color)

    elif h_light == "diag":
        color = base_color

    return color


def _write_tikz(tikz, out_file, build=True, cleanup=True):
    with open('{}.tex'.format(out_file), 'w') as f:
        f.write(base_file_start)
        f.write(tikz)
        f.write(base_file_end)

    if build:
        os.system('pdflatex {}.tex'.format(out_file))

        if cleanup:
            for ext in ['aux', 'fdb_latexmk', 'fls', 'log', 'tex']:
                f_name = '{}.{}'.format(out_file, ext)
                if os.path.exists(f_name):
                    os.remove(f_name)


class TotalJacobian(object):

    def __init__(self):
        self._variables = {}
        self._j_inputs = {}
        self._n_inputs = 0

        self._i_outputs = {}
        self._n_outputs = 0

        self._connections = {}
        self._ij_connections = {}

        self._setup = False

    def add_input(self, name, size=1, text=''):
        self._variables[name] = Variable(
            size=size,
            idx=self._n_inputs,
            text=text,
            color=None)
        self._j_inputs[self._n_inputs] = self._variables[name]
        self._n_inputs += 1

    def add_output(self, name, size=1, text=''):
        self._variables[name] = Variable(
            size=size,
            idx=self._n_outputs,
            text=text,
            color=None)
        self._i_outputs[self._n_outputs] = self._variables[name]
        self._n_outputs += 1

    def connect(self, src, target, text='', color='tableau0'):
        if isinstance(target, (list, tuple)):
            for t in target:
                self._connections[
                    src,
                    t] = CellData(
                    text=text,
                    color=color,
                    highlight='diag')
        else:
            self._connections[
                src,
                target] = CellData(
                text=text,
                color=color,
                highlight='diag')

    def _process_vars(self):

        if self._setup:
            return

        # deal with connections
        for (src, target), cell_data in self._connections.items():
            i_src = self._variables[src].idx
            j_target = self._variables[target].idx

            self._ij_connections[i_src, j_target] = cell_data

        self._setup = True

    def write(self, out_file=None, build=True, cleanup=True):
        """
        Write output files for the matrix equation diagram.  This produces the following:

            - {file_name}.tikz
                A file containing the TIKZ definition of the tikz diagram.
            - {file_name}.tex
                A standalone document wrapped around an include of the TIKZ file which can
                be compiled to a pdf.
            - {file_name}.pdf
                An optional compiled version of the standalone tex file.

        Parameters
        ----------
        file_name : str
            The prefix to be used for the output files
        build : bool
            Flag that determines whether the standalone PDF of the XDSM will be compiled.
            Default is True.
        cleanup: bool
            Flag that determines if padlatex build files will be deleted after build is complete
        """
        self._process_vars()

        tikz = []

        # label the columns
        tikz.append(r'\blockrow{')
        # emtpy column for the row labels
        tikz.append(r'  \blockcol{')
        tikz.append(r'    \blockempty{%s*\comp}{%s*\comp}{%s}\\' % (1, 1, ''))
        tikz.append(r'  }')
        for j in range(self._n_inputs):
            var = self._j_inputs[j]
            col_size = var.size
            tikz.append(r'  \blockcol{')
            tikz.append(
                r'    \blockempty{%s*\comp}{%s*\comp}{%s}\\' %
                (col_size, 1, var.text))
            tikz.append(r'  }')
        tikz.append(r'}')

        for i in range(self._n_outputs):
            output = self._i_outputs[i]
            row_size = output.size

            tikz.append(r'\blockrow{')

            # label the row with the output name
            tikz.append(r'  \blockcol{')
            tikz.append(
                r'    \blockempty{%s*\comp}{%s*\comp}{%s}\\' %
                (1, row_size, output.text))
            tikz.append(r'  }')

            for j in range(self._n_inputs):
                var = self._j_inputs[j]
                col_size = var.size
                tikz.append(r'  \blockcol{')
                if (j, i) in self._ij_connections:
                    cell_data = self._ij_connections[(j, i)]
                    conn_color = 'T{}'.format(var.color)
                    if cell_data.color is not None:
                        conn_color = _color(
                            cell_data.color,
                            cell_data.highlight)
                    tikz.append(
                        r'    \blockmat{%s*\comp}{%s*\comp}{%s}{draw=white,fill=%s}{}\\' %
                        (col_size, row_size, cell_data.text, conn_color))
                else:
                    tikz.append(
                        r'    \blockempty{%s*\comp}{%s*\comp}{}\\' %
                        (col_size, row_size))
                tikz.append(r'  }')

            tikz.append(r'}')

        jac_tikz = "\n".join(tikz)

        _write_tikz(jac_tikz, out_file, build, cleanup)


class MatrixEquation(object):

    def __init__(self):
        self._variables = {}
        self._ij_variables = {}

        self._n_vars = 0

        self._connections = {}
        self._ij_connections = {}

        self._text = {}
        self._ij_text = {}

        self._total_size = 0

        self._setup = False

        self._terms = []

    def clear_terms(self):
        self._terms = []

    def add_variable(self, name, size=1, text='', color='blue'):
        self._variables[name] = Variable(
            size=size,
            idx=self._n_vars,
            text=text,
            color=color)
        self._ij_variables[self._n_vars] = self._variables[name]
        self._n_vars += 1

        self._total_size += size

    def connect(self, src, target, text='', color=None, highlight=1):

        if isinstance(target, (list, tuple)):
            for t in target:
                self._connections[
                    src,
                    t] = CellData(
                    text=text,
                    color=color,
                    highlight=highlight)
        else:
            self._connections[
                src,
                target] = CellData(
                text=text,
                color=color,
                highlight=highlight)

    def text(self, src, target, text):
        "don't connect the src and target, but put some text where a connection would be"
        self._text[src, target] = CellData(text=text, color=None, highlight=-1)

    def _process_vars(self):
        """map all the data onto i,j grid"""

        if self._setup:
            return

        # deal with connections
        for (src, target), cell_data in self._connections.items():
            i_src = self._variables[src].idx
            i_target = self._variables[target].idx

            self._ij_connections[i_src, i_target] = cell_data

        for (src, target), cell_data in self._text.items():
            i_src = self._variables[src].idx
            j_target = self._variables[target].idx

            self._ij_text[i_src, j_target] = cell_data

        self._setup = True

    def jacobian(self, transpose=False):

        self._process_vars()

        tikz = []

        for i in range(self._n_vars):
            tikz.append(r'\blockrow{')

            row_size = self._ij_variables[i].size
            for j in range(self._n_vars):
                var = self._ij_variables[j]
                col_size = var.size
                tikz.append(r'  \blockcol{')

                if transpose:
                    location = (i, j)
                else:
                    location = (j, i)

                if i == j:
                    tikz.append(
                        r'    \blockmat{%s*\comp}{%s*\comp}{%s}{draw=white,fill=D%s}{}\\' %
                        (col_size, row_size, var.text, var.color))
                elif location in self._ij_connections:
                    cell_data = self._ij_connections[location]
                    conn_color = 'T{}'.format(var.color)
                    if cell_data.color is not None:
                        conn_color = _color(
                            cell_data.color,
                            cell_data.highlight)
                    tikz.append(
                        r'    \blockmat{%s*\comp}{%s*\comp}{%s}{draw=white,fill=%s}{}\\' %
                        (col_size, row_size, cell_data.text, conn_color))
                elif location in self._ij_text:
                    cell_data = self._ij_text[location]
                    tikz.append(
                        r'    \blockempty{%s*\comp}{%s*\comp}{%s}\\' %
                        (col_size, row_size, cell_data.text))
                else:
                    tikz.append(
                        r'    \blockempty{%s*\comp}{%s*\comp}{}\\' %
                        (col_size, row_size))
                tikz.append(r'  }')

            tikz.append(r'}')

        lhs_tikz = "\n".join(tikz)

        self._terms.append(lhs_tikz)
        return lhs_tikz

    def vector(self, base_color='red', highlight=None):

        self._process_vars()

        tikz = []

        if highlight is None:
            highlight = np.ones(self._n_vars)

        for i, h_light in enumerate(highlight):

            color = _color(base_color, h_light)

            row_size = self._ij_variables[i].size

            tikz.append(r'\blockrow{\blockcol{')
            if h_light == "diag":
                tikz.append(
                    r'  \blockdiag{1*\comp}{%s*\comp}{}{draw=white,fill=T%s}{\dt}{draw=white,fill=D%s}\\' %
                    (row_size, color, color))
            else:
                tikz.append(
                    r'  \blockmat{1*\comp}{%s*\comp}{}{draw=white,fill=%s}{}\\' %
                    (row_size, color))

            tikz.append(r'}}')

        vec_tikz = "\n".join(tikz)

        self._terms.append(vec_tikz)
        return vec_tikz

    def operator(self, opperator="="):

        self._process_vars()

        tikz = []

        padding_size = (self._total_size - 1) / 2

        tikz.append(r'\blockrow{')
        tikz.append(r'  \blockempty{\mwid}{%s*\comp}{} \\' % (padding_size))
        tikz.append(
            r'  \blockmat{\mwid}{1*\comp}{\huge $%s$}{draw=white,fill=white}{}\\' %
            (opperator))
        tikz.append(r'  \blockempty{\mwid}{%s*\comp}{} \\' % (padding_size))
        tikz.append(r'}')

        op_tikz = "\n".join(tikz)

        self._terms.append(op_tikz)
        return op_tikz

    def spacer(self):

        self._process_vars()

        tikz = []

        for i in range(self._n_vars):
            row_size = self._ij_variables[i].size

            tikz.append(r'\blockrow{\blockcol{')
            tikz.append(
                r'  \blockmat{.25*\mwid}{%s*\comp}{}{draw=white,fill=white}{}\\' %
                (row_size))
            tikz.append(r'}}')

        spacer_tikz = "\n".join(tikz)

        self._terms.append(spacer_tikz)
        return spacer_tikz

    def write(self, out_file=None, build=True, cleanup=True):
        """
        Write output files for the matrix equation diagram.  This produces the following:

            - {file_name}.tikz
                A file containing the TIKZ definition of the tikz diagram.
            - {file_name}.tex
                A standalone document wrapped around an include of the TIKZ file which can
                be compiled to a pdf.
            - {file_name}.pdf
                An optional compiled version of the standalone tex file.

        Parameters
        ----------
        file_name : str
            The prefix to be used for the output files
        build : bool
            Flag that determines whether the standalone PDF of the XDSM will be compiled.
            Default is True.
        cleanup: bool
            Flag that determines if padlatex build files will be deleted after build is complete
        """
        tikz = []
        tikz.append(r'\blockrow{')

        for term in self._terms:
            tikz.append(r'\blockcol{')
            tikz.append(term)
            tikz.append(r'}')
        tikz.append(r'}')

        eqn_tikz = "\n".join(tikz)

        if out_file:
            # with open('{}.tex'.format(out_file), 'w') as f:
            #     f.write(base_file_start)
            #     f.write(eqn_tikz)
            #     f.write(base_file_end)

            # if build:
            #     os.system('pdflatex {}.tex'.format(out_file))

            #     if cleanup:
            #         for ext in ['aux', 'fdb_latexmk', 'fls', 'log', 'tex']:
            #             f_name = '{}.{}'.format(out_file, ext)
            #             if os.path.exists(f_name):
            #                 os.remove(f_name)
            _write_tikz(eqn_tikz, out_file, build, cleanup)


if __name__ == "__main__":
    lst = MatrixEquation()

    lst.add_variable('x', text=r'$x$')
    lst.add_variable('y', size=3, text=r'$y$')
    lst.add_variable('z')

    lst.connect('x', 'y')
    lst.connect('y', 'z')
    lst.text('z', 'x', r'$0$')

    lst.jacobian(transpose=True)
    lst.spacer()
    lst.vector(base_color='green', highlight=[3, 2, 'diag'])
    lst.operator('=')
    lst.vector(base_color='red')
    lst.vector(base_color='red')

    lst.write('test')

    J = TotalJacobian()
    J.add_input('a', text=r'$a$')
    J.add_input('b', text=r'$b$')
    J.add_input('c', text=r'$c$')
    J.add_input('d', text=r'$d$')
    J.add_input('e', text=r'$e$')

    J.add_output('gc', text=r'$g_c$')
    J.add_output('gd', text=r'$g_d$')
    J.add_output('ge', text=r'$g_e$')
    J.add_output('f', text=r'$f$')

    J.connect('a', ('gc', 'gd', 'ge', 'f'))
    J.connect('b', ('gc', 'gd', 'ge', 'f'))
    J.connect('c', 'gc')
    J.connect('d', 'gd')
    J.connect('e', ('ge', 'f'))

    J.write('J_test', cleanup=False)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Francois Gallard, Matthias De Lozzo, Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
K-nearest neighbors classification model
========================================

The k-nearest neighbor classification algorithm is an approach to predict the
output class of a new input point by selecting the majority class
among the k nearest neighbors in a training set through voting. The algorithm
may also predict the probabilties of belonging to each class by counting the
number of occurences of the class withing the k nearest neighbors.

Let :math:`(x_i)_{i=1,\\cdots,n_{\\text{samples}}}\\in
\\mathbb{R}^{n_{\\text{samples}}\\times n_{\\text{inputs}}}` and
:math:`(y_i)_{i=1,\\cdots,n_{\\text{samples}}}\\in
\\{1,\\cdots,n_{\\text{classes}}\\}^{n_{\\text{samples}}}` denote the input and
output training data respectively.

The procedure for predicting the class of a new input point :math:`x\\in
\\mathbb{R}^{n_{\\text{inputs}}}` is the following:

Let :math:`i_1(x), \\cdots, i_{n_{\\text{samples}}}(x)` be the indices
of the input training points sorted by distance to the prediction point
:math:`x`, i.e.

.. math::

    \\|x-x_{i_1(x)}\\| \\leq \\cdots \\leq
    \\|x-x_{i_{n_{\\text{samples}}}(x)}\\|.

The ordered indices may be formally determined through the inductive formula

.. math::

    i_p(x) = \\underset{i\\in I_p(x)}{\\operatorname{argmin}}\\|x-x_i\\|,\\quad
    p=1,\\cdots,n_{\\text{samples}}

where

.. math::

    I_1(x) = \\{1,\\cdots,n_{\\text{samples}}\\}\\\\
    I_{p+1} = I_p(x)\\setminus \\{i_p(x)\\},\\quad
    p=1,\\cdots,n_{\\text{samples}}-1,

that is

.. math::

    I_p(x) = \\{1,\\cdots,n_{\\text{samples}}\\}\\setminus
    \\{i_1(x),\\cdots,i_{p-1}(x)\\}.

Then, by denoting :math:`\\operatorname{mode}(\\cdot)` the mode operator, i.e.
the operator that extracts the element with the highest occurence,
we may define the prediction operator as the mode of the set of output classes
associated to the :math:`k` first indices (classes of the :math:`k`-nearest
neighbors of :math:`x`):

.. math::

    f(x) = \\operatorname{mode}(y_{i_1(x)}, \\cdots, y_{i_k(x)})


This concept is implemented through the :class:`.KNNClassifier` class which
inherits from the :class:`.MLClassificationAlgo` class.

Dependence
----------
The classifier relies on the KNeighborsClassifier class
of the `scikit-learn library <https://scikit-learn.org/stable/modules/
generated/sklearn.neighbors.KNeighborsClassifier.html>`_.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from numpy import stack
from sklearn.neighbors import KNeighborsClassifier as SKLKNN

from gemseo.mlearning.classification.classification import MLClassificationAlgo

standard_library.install_aliases()


from gemseo import LOGGER


class KNNClassifier(MLClassificationAlgo):
    """ K nearest neighbors classification algorithm. """

    LIBRARY = "scikit-learn"
    ABBR = "KNN"

    def __init__(
        self,
        data,
        transformer=None,
        input_names=None,
        output_names=None,
        n_neighbors=5,
        **parameters
    ):
        """Constructor.

        :param data: learning dataset.
        :type data: Dataset
        :param transformer: transformation strategy for data groups.
            If None, do not transform data. Default: None.
        :type transformer: dict(str)
        :param input_names: names of the input variables.
        :type input_names: list(str)
        :param output_names: names of the output variables.
        :type output_names: list(str)
        :param n_neighbors: number of neighbors.
        :type n_neighbords: int
        :param parameters: other keyword arguments for sklearn KNN.
        """
        super(KNNClassifier, self).__init__(
            data,
            transformer=transformer,
            input_names=input_names,
            output_names=output_names,
            n_neighbors=n_neighbors,
            **parameters
        )
        self.algo = SKLKNN(n_neighbors, **parameters)

    def _fit(self, input_data, output_data):
        """Fit the classification model.

        :param ndarray input_data: input data (2D).
        :param ndarray(int) output_data: output data.
        """
        if output_data.shape[1] == 1:
            output_data = output_data.ravel()
        self.algo.fit(input_data, output_data)

    def _predict(self, input_data):
        """Predict output data from input data.

        :param ndarray input_data: input data (n_samples, n_inputs).
        :return: output data (n_samples, n_outputs).
        :rtype: ndarray(int)
        """
        output_data = self.algo.predict(input_data).astype(int)
        if len(output_data.shape) == 1:
            output_data = output_data[:, None]
        return output_data

    def _predict_proba_soft(self, input_data):
        """Predict probability of belonging to each class.

        :param ndarray input_data: input data (n_samples, n_inputs).
        :return: probabilities of belonging to each class
            (n_samples, n_outputs, n_classes). For a given sample and output
            variable, the sum of probabilities is one.
        :rtype: ndarray
        """
        probas = self.algo.predict_proba(input_data)
        if len(probas[0].shape) == 1:
            probas = probas[..., None]
        else:
            probas = stack(probas, axis=-1)
        return probas



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Mixture of Experts
==================

The mixture of experts (MoE) regression model expresses the output as a
weighted sum of local surrogate models, where the weights are indicating
the class of the input.

Inputs are grouped into clusters by a classification model that is trained on a
training set where the output labels are determined through a clustering
algorithm. The outputs may be preprocessed trough a sensor or a dimension
reduction algorithm.

The classification may either be hard, in which only one of the weights is
equal to one, and the rest equal to zero:

.. math::

    y = \\sum_{k=1}^K i_{C_k}(x) f_k(x),

or soft, in which case the weights express the probabilities of belonging to
each class:

.. math::

    y = \\sum_{k=1}^K \\mathbb{P}(x\\in C_k) f_k(x),

where
:math:`x` is the input,
:math:`y` is the output,
:math:`K` is the number of classes,
:math:`(C_k)_{k=1,\\cdots,K}` are the input spaces associated to the classes,
:math:`i_{C_k}(x)` is the indicator of class :math:`k`,
:math:`\\mathbb{P}(x\\in C_k)` is the probability of class :math:`k`
given :math:`x` and
:math:`f_k(x)` is the local surrogate model on class :math:`k`.

This concept is implemented through the :class:`.MixtureOfExperts` class which
inherits from the :class:`.MLRegressionAlgo` class.
"""
from __future__ import absolute_import, division, unicode_literals

from os.path import join

from future import standard_library
from numpy import nonzero, unique, where, zeros

from gemseo.core.dataset import Dataset
from gemseo.mlearning.classification.factory import ClassificationModelFactory
from gemseo.mlearning.cluster.factory import ClusteringModelFactory
from gemseo.mlearning.regression.factory import RegressionModelFactory
from gemseo.mlearning.regression.regression import MLRegressionAlgo
from gemseo.utils.data_conversion import DataConversion

standard_library.install_aliases()


from gemseo import LOGGER


class MixtureOfExperts(MLRegressionAlgo):
    """ Mixture of experts regression. """

    ABBR = "MoE"

    LABELS = "labels"

    _LOCAL_INPUT = "input"
    _LOCAL_OUTPUT = "output"

    def __init__(
        self, data, transformer=None, input_names=None, output_names=None, hard=True
    ):
        """Constructor.

        :param data: learning dataset.
        :type data: Dataset
        :param transformer: transformation strategy for data groups.
            If None, do not transform data. Default: None.
        :type transformer: dict(str)
        :param input_names: names of the input variables.
        :type input_names: list(str)
        :param output_names: names of the output variables.
        :type output_names: list(str)
        :param hard: Indicator for hard or soft clustering/classification.
            Hard clustering/classification if True. Default: True.
        :type hard: bool
        """
        super(MixtureOfExperts, self).__init__(
            data,
            transformer=transformer,
            input_names=input_names,
            output_names=output_names,
            hard=hard,
        )
        self.hard = hard
        self.cluster_algo = "KMeans"
        self.classif_algo = "KNNClassifier"
        self.regress_algo = "LinearRegression"
        self.cluster_params = {}
        self.classif_params = {}
        self.regress_params = {}

        self.clusterer = None
        self.classifier = None
        self.regress_models = None

    class DataFormatters(MLRegressionAlgo.DataFormatters):
        """ Machine learning regression model decorators. """

        @classmethod
        def format_predict_class_dict(cls, predict):
            """If input_data is passed as a dictionary, then convert it to
            ndarray, and convert output_data to dictionary. Else, do nothing.

            :param predict: Method whose input_data and output_data are to be
                formatted.
            """

            def wrapper(self, input_data, *args, **kwargs):
                """ Wrapper function. """
                as_dict = isinstance(input_data, dict)
                if as_dict:
                    input_data = DataConversion.dict_to_array(
                        input_data, self.input_names
                    )
                output_data = predict(self, input_data, *args, **kwargs)
                if as_dict:
                    output_data = {self.LABELS: output_data}
                return output_data

            return wrapper

    def set_clusterer(self, cluster_algo, **cluster_params):
        """Set cluster algorithm.

        :param str cluster_algo: clusterer.
        :param cluster_params: optional arguments for clustering.
            If none, uses default arguments.
        """
        self.cluster_algo = cluster_algo
        self.cluster_params = cluster_params

    def set_classifier(self, classif_algo, **classif_params):
        """Set classification algorithm.

        :param str classif_algo: classifier.
        :param classif_params: optional arguments for classification.
            If none, uses default arguments.
        """
        self.classif_algo = classif_algo
        self.classif_params = classif_params

    def set_regressor(self, regress_algo, **regress_params):
        """Set regression algorithm.

        :param str regress_algo: regressor.
        :param regress_params: optional arguments for regression.
            If none, uses default arguments.
        """
        self.regress_algo = regress_algo
        self.regress_params = regress_params

    @DataFormatters.format_predict_class_dict
    @DataFormatters.format_samples
    @DataFormatters.format_transform(transform_outputs=False)
    def predict_class(self, input_data):
        """Predict classes of input data.

        :param input_data: input data (1D or 2D).
        :type input_data: dict(ndarray) or ndarray
        :return: output classes ("0D" or 1D, one less than input data).
        :rtype: int or ndarray(int)
        """
        return self.classifier.predict(input_data)

    @DataFormatters.format_input_output
    def predict_local_model(self, input_data, index):
        """Predict output for given input data, using an individual regression
        model from the list.

        :param input_data: input data (1D or 2D).
        :type input_data: dict(ndarray) or ndarray
        :param int index: index of the local regression model.
        :return: output data (1D or 2D, same as input_data).
        :rtype: dict(ndarray) or ndarray
        """
        return self.regress_models[index].predict(input_data)

    def _fit(self, input_data, output_data):
        """Fit the regression model. As the data is provided as two numpy
        arrays, we construct a temporary dataset in order to use clustering,
        classification and regression algorithms.

        :param ndarray input_data: input data (2D).
        :param ndarray output_data: output data (2D).
        """
        dataset = Dataset("training_set")
        dataset.add_group(
            Dataset.INPUT_GROUP,
            input_data,
            [self._LOCAL_INPUT],
            {self._LOCAL_INPUT: input_data.shape[1]},
        )
        dataset.add_group(
            Dataset.OUTPUT_GROUP,
            output_data,
            [self._LOCAL_OUTPUT],
            {self._LOCAL_OUTPUT: output_data.shape[1]},
            cache_as_input=False,
        )
        self._fit_clusters(dataset)
        self._fit_classifier(dataset)
        self._fit_regressors(dataset)

    def _fit_clusters(self, dataset):
        """Fit the clustering algorithm to the dataset (input/output labels
        ignored by clustering algorithm). Add resulting labels as a new output
        in the dataset.

        :param Dataset dataset: dataset containing input and output data.
        """
        factory = ClusteringModelFactory()
        self.clusterer = factory.create(
            self.cluster_algo, data=dataset, **self.cluster_params
        )
        self.clusterer.learn()
        labels = self.clusterer.labels[:, None]
        dataset.add_variable(self.LABELS, labels, self.LABELS, False)

    def _fit_classifier(self, dataset):
        """Fit the input data to the labels using the given classification
        algorithm.

        :param Dataset dataset: dataset containing input and output data,
            as well as labels after execution of _fit_clusters().
        """
        factory = ClassificationModelFactory()
        self.classifier = factory.create(
            self.classif_algo,
            data=dataset,
            output_names=[self.LABELS],
            **self.classif_params
        )
        self.classifier.learn()

    def _fit_regressors(self, dataset):
        """Fit the local regression models on each cluster separately.

        :param Dataset dataset: dataset containing input and output data.
        """
        key = "transformer"
        transform = False
        transformer = {}
        if key in self.regress_params:
            transformer_template = self.regress_params.pop(key)
            transform = True

        factory = RegressionModelFactory()

        self.regress_models = []
        for index in range(self.clusterer.n_clusters):
            samples = nonzero(self.clusterer.labels == index)[0]
            if transform:
                transformer = {
                    name: val.duplicate() for name, val in transformer_template.items()
                }
            local_model = factory.create(
                self.regress_algo,
                data=dataset,
                transformer=transformer,
                **self.regress_params
            )
            local_model.learn(samples=samples)
            self.regress_models.append(local_model)

    def _predict_all(self, input_data):
        """Predict output of each regression model for given input data. Stack
        the different outputs along a new axis.

        :param ndarray input_data: input data (2D).
        :return: all output predictions (3D).
        :rtype: ndarray
        """
        # dim(input_data)  = (n_samples, n_inputs)
        # dim(output_data) = (n_samples, n_clusters, n_outputs)
        output_data = zeros(
            (input_data.shape[0], self.n_clusters, self.regress_models[0].output_shape)
        )
        for i in range(self.n_clusters):
            output_data[:, i] = self.regress_models[i].predict(input_data)
        return output_data

    def _predict(self, input_data):
        """Predict global output for given input data. The global output is
        computed as a sum of contributions from the individual local regression
        models, weighted by the probabilities of belonging to each cluster.

        :param ndarray input_data: input data (2D).
        :param bool hard: indicator for hard or soft classification.
            Default: True
        :return: global output prediction (2D).
        :rtype: ndarray
        """
        # dim(probas)         = (n_samples, n_clusters,     1    )
        # dim(local_outputs)  = (n_samples, n_clusters, n_outputs)
        # dim(contributions)  = (n_samples, n_clusters, n_outputs)
        # dim(global_outputs) = (n_samples, n_outputs)
        probas = self.classifier.predict_proba(input_data, hard=self.hard)
        local_outputs = self._predict_all(input_data)
        contributions = probas * local_outputs
        global_outputs = contributions.sum(axis=1)
        return global_outputs

    def _predict_jacobian(self, input_data):
        """Predict Jacobian of the regression model for the given input data.

        :param ndarray input_data: input_data (2D).
        :return: Jacobian matrices (3D, one for each sample).
        :rtype: ndarray
        """
        if self.hard:
            jacobians = self._predict_jacobian_hard(input_data)
        else:
            jacobians = self._predict_jacobian_soft(input_data)
        return jacobians

    def _predict_jacobian_hard(self, input_data):
        """Predict Jacobian of the regression model for the given input data,
        with a hard (constant) classification.

        :param ndarray input_data: input_data (2D).
        :return: Jacobian matrices (3D, one 2D matrix for each sample).
        :rtype: ndarray
        """
        n_samples = input_data.shape[0]
        classes = self.classifier.predict(input_data)[..., 0]
        unq_classes = unique(classes)
        jacobians = zeros(
            (
                n_samples,
                self.regress_models[0].output_shape,
                self.regress_models[0].input_shape,
            )
        )
        for klass in unq_classes:
            inds_kls = where(classes == klass)
            jacobians[inds_kls] = self.regress_models[klass].predict_jacobian(
                input_data[inds_kls]
            )
        return jacobians

    def _predict_jacobian_soft(self, input_data):
        """Predict Jacobian of the regression model for the given input data,
        with a soft classification.

        :param ndarray input_data: input_data (2D).
        :return: Jacobian matrices (3D, one 2D matrix for each sample).
        :rtype: ndarray
        """
        raise NotImplementedError

    def _save_algo(self, directory):
        """Save external machine learning algorithm.

        :param str directory: algorithm directory.
        """
        self.clusterer.save(join(directory, "clusterer"))
        self.classifier.save(join(directory, "classifier"))
        for i, local_model in enumerate(self.regress_models):
            local_model.save(join(directory, "local_model_{}".format(i)))

    def load_algo(self, directory):
        """Load external machine learning algorithm.

        :param str directory: algorithm directory.
        """
        cluster_factory = ClusteringModelFactory()
        classif_factory = ClassificationModelFactory()
        regress_factory = RegressionModelFactory()
        self.clusterer = cluster_factory.load(join(directory, "clusterer"))
        self.classifier = classif_factory.load(join(directory, "classifier"))
        self.regress_models = []
        for i in range(self.n_clusters):
            self.regress_models.append(
                regress_factory.load(join(directory, "local_model_{}".format(i)))
            )

    def __str__(self):
        """ String representation for end user. """
        string = super(MixtureOfExperts, self).__str__()
        string += "\n\nClusterer:\n{}".format(self.clusterer)
        string += "\n\nClassifier:\n{}".format(self.classifier)
        for i, local_model in enumerate(self.regress_models):
            string += "\n\nLocal model {}:\n{}".format(i, local_model)
        return string

    def _get_objects_to_save(self):
        """Get objects to save.
        :return: objects to save.
        :rtype: dict
        """
        objects = super(MixtureOfExperts, self)._get_objects_to_save()
        objects["cluster_algo"] = self.cluster_algo
        objects["classif_algo"] = self.classif_algo
        objects["regress_algo"] = self.regress_algo
        objects["cluster_params"] = self.cluster_params
        objects["classif_params"] = self.classif_params
        objects["regress_params"] = self.regress_params
        return objects

    @property
    def labels(self):
        """ Cluster labels. """
        return self.clusterer.labels

    @property
    def n_clusters(self):
        """ Number of clusters. """
        return self.clusterer.n_clusters



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                        documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Python logger configuration made easier
***************************************
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import sys

from future import standard_library
from future.utils import with_metaclass

from gemseo.utils.singleton import SingleInstancePerAttributeId

standard_library.install_aliases()


class LoggerConfig(with_metaclass(SingleInstancePerAttributeId, object)):
    """Class to easily set the self.logger configuration."""

    def __init__(self, logger):
        """
        Initialize handlers
        """
        self.logger = logger
        logger_streams = [
            hdlr.stream for hdlr in logger.handlers if hasattr(hdlr, "stream")
        ]
        if sys.stdout not in logger_streams:
            self.stdout_hdlr = logging.StreamHandler(sys.stdout)
            self.logger.addHandler(self.stdout_hdlr)
        else:
            hdlr_indx = logger_streams.index(sys.stdout)
            self.stdout_hdlr = logger.handlers[hdlr_indx]

        if sys.stderr not in logger_streams:
            self.stderr_hdlr = logging.StreamHandler(sys.stderr)
            self.logger.addHandler(self.stderr_hdlr)
        else:
            hdlr_indx = logger_streams.index(sys.stderr)
            self.stderr_hdlr = logger.handlers[hdlr_indx]
        self.file_handler = None
        self.formatter = None
        self.logger.propagate = False

    def activate_debug(self):
        """Activate the debug prints"""
        self.logger.setLevel(logging.DEBUG)

    def deactivate_debug(self):
        """De Activate the debug prints, show info or more critical"""
        self.logger.setLevel(logging.INFO)

    def hide_non_warnings(self):
        """Activate the warning and critical prints"""
        self.logger.setLevel(logging.WARNING)

    def show_info(self):
        """Shows info or more critical"""
        self.logger.setLevel(logging.INFO)

    def set_level(self, level):
        """Set logging level

        :param level: logger level (DEBUG, INFO...)
        """
        self.logger.setLevel(level)

    def add_logging_file(self, filename="mdo_scenario.log", mode="a", delay=True):
        """Adds a logging file

        :param filename: the output file (Default value = "mdo_scenario.log")
        :param mode: write mode of the file (Default value = "a")
        :param delay: if True , waits for the first emit to write the log file
            (Default value = True)
        """
        self.file_handler = logging.FileHandler(
            filename, mode=mode, delay=delay, encoding="utf-8"
        )
        self.logger.addHandler(self.file_handler)
        self.file_handler.setFormatter(self.formatter)

    def deactivate_file_logging(self):
        """Deactivate file logging"""
        self.logger.removeHandler(self.file_handler)

    def set_logger_config(
        self,
        level=None,
        date_format=None,
        message_format=None,
        filename=None,
        filemode="a",
    ):
        """Sets the self.logger configuration

        :param level: self.logger print level, default INFO, can be :
            self.logger.DEBUG, self.logger.INFO,
            self.logger.WARNING, logging.CRITICAL
        :param date_format: date format, if None, use a default one
        :param message_format: message format, if None, use a default one
        :param filename: the file path if outputs must be written in a file
            (Default value = None)
        :param filemode: Default value = 'a')
        """
        if level is None:
            level = logging.INFO
        if date_format is None:
            date_format = "%d/%m/%Y %H:%M:%S"
            date_format = "%H:%M:%S"
        if message_format is None:
            message_format = "%(levelname)8s - %(asctime)s : %(message)s"
        self.formatter = logging.Formatter(message_format, date_format)
        if filename is not None:
            self.add_logging_file(filename, mode=filemode, delay=True)

        self.logger.setLevel(level)
        self.stdout_hdlr.addFilter(MaxLevelFilter(logging.INFO))
        self.stdout_hdlr.setLevel(level)
        self.stderr_hdlr.setLevel(logging.WARNING)

        self.stdout_hdlr.setFormatter(self.formatter)
        self.stderr_hdlr.setFormatter(self.formatter)

        # Capture warnings module outputs, used in matplotlib for instance
        logging.captureWarnings(True)
        self.logger.gemseo_config = self


class MaxLevelFilter(logging.Filter):
    """Filters (lets through) all messages with level <= LEVEL"""

    def __init__(self, level):
        """
        Constructor

        :param level: max message level
        """
        self.level = level
        super(MaxLevelFilter, self).__init__()

    def filter(self, record):
        """Filters a log message if the level of the message
        is lower than self.level

        :param record: the log record
        :returns: True if the message should be displayed

        """
        return record.levelno <= self.level



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Data transformer
================

The abstract :class:`.Transformer` class implements the concept of a data
transformer. Inheriting classes should implement the
:meth:`.Transformer.fit`, :meth:`.Transformer.transform` and
possibly :meth:`.Transformer.inverse_transform` methods.

.. seealso::

   :mod:`~gemseo.mlearning.transform.scaler.scaler`
   :mod:`~gemseo.mlearning.transform.dimension_reduction.dimension_reduction`
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library

standard_library.install_aliases()


class Transformer(object):
    """ Transformer baseclass. """

    def __init__(self, name="Transformer", **parameters):
        """Constructor.

        :param str name: transformer name. Default: 'Transformer'.
        :param parameters: transformer parameters.
        """
        self.name = name
        self.parameters = parameters

    def duplicate(self):
        """ Duplicate the constructor. """
        return self.__class__(self.name, **self.parameters)

    def fit(self, data):
        """Fit transformer to data.

        :param ndarray data: data to be fitted.
        """
        raise NotImplementedError

    def transform(self, data):
        """Transform data.

        :param ndarray data: data  to be transformed.
        :return: transformed data.
        :rtype: ndarray
        """
        raise NotImplementedError

    def inverse_transform(self, data):
        """Perform an inverse transform on the data.

        :param ndarray data: data to be inverse transformed.
        :return: inverse transformed data.
        :rtype: ndarray
        """
        raise NotImplementedError

    def fit_transform(self, data):
        """Fit transformer to data and transform data.

        :param ndarray data: data to be fitted and transformed.
        :return: transformed data.
        :rtype: ndarray
        """
        self.fit(data)
        transformed_data = self.transform(data)
        return transformed_data

    def compute_jacobian(self, data):
        """Compute Jacobian of the transformer transform.

        :param ndarray data: data where the Jacobian is to be computed.
        :return: Jacobian matrix.
        :rtype: ndarray
        """
        raise NotImplementedError

    def compute_jacobian_inverse(self, data):
        """Compute Jacobian of the transformer inverse_transform.

        :param ndarray data: data where the Jacobian is to be computed.
        :return: Jacobian matrix.
        :rtype: ndarray
        """
        raise NotImplementedError

    def __str__(self):
        """ String representation for end user. """
        string = self.__class__.__name__
        return string



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Pierre-Jean Barjhoux
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
A radar plot of constraints
***************************
"""
from __future__ import absolute_import, division, unicode_literals

from matplotlib import pyplot
from numpy import concatenate, linspace, max, pi, rad2deg, zeros

from gemseo import LOGGER
from gemseo.post.opt_post_processor import OptPostProcessor


class RadarChart(OptPostProcessor):
    """Plot on radar style chart a list of constraint functions.

    This class has the responsability of plotting on radar style chart a list
    of constraint functions at a given iteration.

    By default, the iteration is the last one.
    It is possible either to save the plot, to show the plot or both.
    """

    def _plot(
        self,
        constraints_list,
        iteration=-1,
        show=False,
        save=False,
        file_path="radar_chart",
        extension="pdf",
        figsize_x=8,
        figsize_y=8,
    ):
        """Plot radar graph.

        :param constraints_list: list of constraints names
        :type constraints_list: list(str)
        :param iteration: number of iteration to post process
        :type iteration: int
        :param show: if True, displays the plot windows
        :type show: bool
        :param save: if True, exports plot to pdf
        :type save: bool
        :param file_path: the base paths of the files to export
        :type file_path: str
        :param extension: file extension
        :type extension: str
        """
        # retrieve the constraints values
        add_dv = False
        all_constr_names = self.opt_problem.get_constraints_names()

        for func in constraints_list:
            if func not in all_constr_names:
                raise ValueError(
                    "Cannot build radar chart,"
                    + " Function "
                    + func
                    + " is not among constraints names"
                    + " or does not exist."
                )

        vals, vname, _ = self.database.get_history_array(
            constraints_list, add_dv=add_dv
        )

        if iteration < -1 or iteration >= len(self.database):
            raise ValueError(
                "iteration should be positive and lower than"
                + " maximum iteration ="
                + str(len(self.database))
            )
        cstr_values = vals[iteration, :]

        # formating values, max and min
        values = cstr_values.ravel()
        vmax = max(cstr_values)
        vmin = min(cstr_values)

        # radar solid grid lines
        pyplot.rc("grid", color="k", linewidth=0.3, linestyle=":")
        pyplot.rc("xtick", labelsize=10)
        pyplot.rc("ytick", labelsize=10)

        # force square figure and square axes looks better for polar, IMO
        figsize = (figsize_x, figsize_y)
        fig = pyplot.figure(figsize=figsize)
        axe = fig.add_axes([0.1, 0.1, 0.8, 0.8], projection="polar")
        constr_nb = len(vname)
        constr_trigger = zeros(constr_nb)

        # computes angles
        theta = 2 * pi * linspace(0, 1 - 1.0 / constr_nb, constr_nb)

        # plot lines
        lines_cstr_comp = axe.plot(
            theta, values, color="k", lw=1, label="computed constraints"
        )
        self.__close_line(lines_cstr_comp[0])

        lines_cstr_lim = axe.plot(
            theta,
            constr_trigger,
            color="red",
            ls="--",
            lw=1,
            label="maximum constraint",
        )
        self.__close_line(lines_cstr_lim[0])

        # display settings
        theta_degree = rad2deg(theta)
        axe.set_thetagrids(theta_degree, vname)
        axe.set_rlim([vmin, vmax])
        axe.set_rticks(linspace(vmin, vmax, 6))
        # set title
        if iteration == -1:
            title = "Constraints at last iteration"
        else:
            title = "Constraints at iteration %i" % iteration
        axe.set_title(title)

        # Shrink current axis's height by 10% on the bottom
        box = axe.get_position()
        axe.set_position(
            [box.x0, box.y0 + box.height * 0.15, box.width, box.height * 0.85]
        )

        axe.legend(loc="upper center", bbox_to_anchor=(0.5, -0.15))

        # save or show
        self._save_and_show(
            fig, save=save, show=show, file_path=file_path, extension=extension
        )

    @staticmethod
    def __close_line(line):
        """
        Closes line plotted in the radar chart

        :param : line (matplotlib oject) to close
        """
        x_values, y_values = line.get_data()
        x_values = concatenate((x_values, [x_values[0]]))
        y_values = concatenate((y_values, [y_values[0]]))
        line.set_data(x_values, y_values)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Surface plot
============

A :class:`.Surfaces` plot represents samples of a functional variable
:math:`z(x,y)` discretized over a 2D mesh. Both evaluations of :math:`z`
and mesh are stored in a :class:`.Dataset`, :math:`z` as a parameter
and the mesh as a metadata.
"""
from __future__ import absolute_import, division, unicode_literals

import matplotlib.pyplot as plt
import matplotlib.tri as mtri
from future import standard_library

from gemseo.post.dataset.dataset_plot import DatasetPlot

standard_library.install_aliases()


class Surfaces(DatasetPlot):
    """ Plot surfaces y_i over the mesh x. """

    def _plot(
        self,
        mesh,
        variable,
        samples=None,
        colormap="Blues",
        add_points=False,
        color="blue",
    ):
        """Curve.

        :param mesh: name of the mesh stored in Dataset.metadata.
        :type mesh: str
        :param variable: variable name for the x-axis.
        :type variable: float
        :param samples: samples indices. If None, plot all samples.
            Default: None.
        :type samples: list(int)
        :param colormap: colormap. Default: 'Blues'.
        :type color: str
        :param add_points: display points over the surface plot.
            Default: False.
        :type add_points: bool
        :param color: point color. Default: blue.
        :type color: str
        """
        x_data = self.dataset.metadata[mesh][:, 0]
        y_data = self.dataset.metadata[mesh][:, 1]
        if samples is not None:
            outputs = self.dataset[variable][variable][samples, :]
        else:
            outputs = self.dataset[variable][variable]

        sample = 0
        fig = []
        for z_data in outputs:
            fig.append(plt.figure())
            axes = fig[-1].add_subplot(1, 1, 1)
            triang = mtri.Triangulation(x_data, y_data)
            tcf = axes.tricontourf(triang, z_data, cmap=colormap)
            if add_points:
                axes.scatter(x_data, y_data, color=color)
            axes.set_title(variable)
            fig[-1].colorbar(tcf)
            fig[-1] = plt.gcf()
            sample += 1
        return fig



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                        documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
SSBJ Disciplines wrappers
*************************
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import super

from future import standard_library
from numpy import array

from gemseo.core.discipline import MDODiscipline
from gemseo.problems.sobieski.core import SobieskiProblem

standard_library.install_aliases()


DTYPE_COMPLEX = "complex128"
DTYPE_DOUBLE = "float64"


class SobieskiBaseWrapperSimpleGram(MDODiscipline):
    """Base wrapper for Sobieski problem discipline wrappers
    and SimpleGrammar

    """

    def __init__(self, dtype=DTYPE_DOUBLE):
        """
        Constructor

        :param dtype: type of data, either "float64" or "complex128"
        :type dtype: str
        """
        super(SobieskiBaseWrapperSimpleGram, self).__init__(
            auto_detect_grammar_files=False,
            grammar_type=MDODiscipline.SIMPLE_GRAMMAR_TYPE,
        )
        self.sobieski_problem = SobieskiProblem(dtype=dtype)
        self.init_values = {}
        self.dtype = dtype

    def _auto_set_all_numpy_types(self):
        """Sets the data types of grammars to numpy array"""
        arr_type = type(array([0.0], dtype=self.dtype))
        n_inputs = len(self.input_grammar.get_data_names())
        n_outputs = len(self.output_grammar.get_data_names())
        self.input_grammar.data_types = [arr_type] * n_inputs
        self.output_grammar.data_types = [arr_type] * n_outputs

    def _run(self):
        """
        Run the discipline
        """
        raise NotImplementedError()


class SobieskiMissionSG(SobieskiBaseWrapperSimpleGram):

    """Sobieski range wrapper using Breguet formula
    Uses simple grammars and not JSON, mainly for concept proof
    Please use JSON version with enhanced checks and features


    """

    def __init__(self, dtype=DTYPE_DOUBLE):
        """
        Constructor of wrapper for weight computation

        :param dtype: type of data, either "float64" or "complex128"
        :type dtype: str
        """
        super(SobieskiMissionSG, self).__init__(dtype=dtype)
        self.input_grammar.data_names = ["y_14", "y_24", "y_34", "x_shared"]
        self.output_grammar.data_names = ["y_4"]
        self._auto_set_all_numpy_types()

    def _run(self):
        """Compute range"""
        data_names = ["y_14", "y_24", "y_34", "x_shared"]
        y_14, y_24, y_34, x_shared = self.get_inputs_by_name(data_names)
        y_4 = self.sobieski_problem.blackbox_mission(x_shared, y_14, y_24, y_34)
        self.store_local_data(y_4=y_4)

    def _compute_jacobian(self, inputs=None, outputs=None):
        """Linearization of weight analysis

        :param inputs: Default value = None)
        :param outputs: Default value = None)
        """
        data_names = ["y_14", "y_24", "y_34", "x_shared"]
        y_14, y_24, y_34, x_shared = self.get_inputs_by_name(data_names)
        self.jac = self.sobieski_problem.derive_blackbox_mission(
            x_shared, y_14, y_24, y_34
        )


class SobieskiStructureSG(SobieskiBaseWrapperSimpleGram):
    """Sobieski mass estimation wrapper
    Uses simple grammars and not JSON, mainly for concept proof
    Please use JSON version with enhanced checks and features
    """

    def __init__(self, dtype=DTYPE_DOUBLE):
        """
        Constructor of wrapper for weight computation
        """
        super(SobieskiStructureSG, self).__init__(dtype=dtype)
        self.input_grammar.data_names = ["x_1", "y_21", "y_31", "x_shared"]
        self.output_grammar.data_names = ["y_1", "y_11", "y_12", "y_14", "g_1"]
        self._auto_set_all_numpy_types()

    def _run(self):
        """Compute weight"""
        data_names = ["x_shared", "y_21", "y_31", "x_1"]
        x_shared, y_21, y_31, x_1 = self.get_inputs_by_name(data_names)
        y_1, y_11, y_12, y_14, g_1 = self.sobieski_problem.blackbox_structure(
            x_shared, y_21, y_31, x_1
        )
        self.store_local_data(y_1=y_1, y_11=y_11, y_12=y_12, y_14=y_14, g_1=g_1)

    def _compute_jacobian(self, inputs=None, outputs=None):
        """Linearization of weight analysis

        :param inputs: Default value = None)
        :param outputs: Default value = None)
        """
        data_names = ["x_shared", "y_21", "y_31", "x_1"]
        x_shared, y_21, y_31, x_1 = self.get_inputs_by_name(data_names)
        self.jac = self.sobieski_problem.derive_blackbox_structure(
            x_shared, y_21, y_31, x_1
        )


class SobieskiAerodynamicsSG(SobieskiBaseWrapperSimpleGram):

    """Sobieski aerodynamic discipline wrapper
    Uses simple grammars and not JSON, mainly for concept proof
    Please use JSON version with enhanced checks and features
    """

    def __init__(self, dtype=DTYPE_DOUBLE):
        """
        Constructor of wrapper for aerodynamic computation

        :param dtype: type of data, either "float64" or "complex128"
        :type dtype: str
        """
        super(SobieskiAerodynamicsSG, self).__init__(dtype=dtype)
        self.input_grammar.data_names = ["x_2", "y_12", "y_32", "x_shared"]
        self.output_grammar.data_names = ["y_2", "y_21", "y_23", "y_24", "g_2"]
        self._auto_set_all_numpy_types()

    def _run(self):
        """Compute aerodynamics"""
        data_names = ["x_2", "y_12", "y_32", "x_shared"]
        x_2, y_12, y_32, x_shared = self.get_inputs_by_name(data_names)
        y_2, y_21, y_23, y_24, g_2 = self.sobieski_problem.blackbox_aerodynamics(
            x_shared, y_12, y_32, x_2
        )
        self.store_local_data(y_2=y_2, y_21=y_21, y_23=y_23, y_24=y_24, g_2=g_2)

    def _compute_jacobian(self, inputs=None, outputs=None):
        """Compute the partial derivatives of all outputs wrt all inputs

        :param inputs: Default value = None)
        :param outputs: Default value = None)
        """
        data_names = ["x_2", "y_12", "y_32", "x_shared"]
        x_2, y_12, y_32, x_shared = self.get_inputs_by_name(data_names)
        self.jac = self.sobieski_problem.derive_blackbox_aerodynamics(
            x_shared, y_12, y_32, x_2
        )


class SobieskiPropulsionSG(SobieskiBaseWrapperSimpleGram):
    """Sobieski propulsion discipline wrapper
    Uses simple grammars and not JSON, mainly for concept proof
    Please use JSON version with enhanced checks and features
    """

    def __init__(self, dtype=DTYPE_DOUBLE):
        """
        Constructor of wrapper for propulsion computation

        :param dtype: type of data, either "float64" or "complex128"
        :type dtype: str
        """
        super(SobieskiPropulsionSG, self).__init__(dtype=dtype)
        self.input_grammar.data_names = ["x_3", "y_23", "x_shared"]
        self.output_grammar.data_names = ["y_3", "y_34", "y_31", "y_32", "g_3"]
        self._auto_set_all_numpy_types()

    def _run(self):
        """Compute propulsion"""
        data_names = ["x_3", "y_23", "x_shared"]
        x_3, y_23, x_shared = self.get_inputs_by_name(data_names)
        y_3, y_34, y_31, y_32, g_3 = self.sobieski_problem.blackbox_propulsion(
            x_shared, y_23, x_3
        )
        self.store_local_data(y_3=y_3, y_34=y_34, y_31=y_31, y_32=y_32, g_3=g_3)

    def _compute_jacobian(self, inputs=None, outputs=None):
        """Compute the partial derivatives of all outputs wrt all inputs

        :param inputs: Default value = None)
        :param outputs: Default value = None)
        """
        data_names = ["x_3", "y_23", "x_shared"]
        x_3, y_23, x_shared = self.get_inputs_by_name(data_names)
        self.jac = self.sobieski_problem.derive_blackbox_propulsion(x_shared, y_23, x_3)
